import os
import warnings
from pathlib import Path
import numpy as np
from dm_control import mjcf
import xml.etree.ElementTree as ET
from mushroom_rl.utils.mujoco import *

import mujoco
import loco_mujoco
from loco_mujoco.environments import ValidTaskConf
from loco_mujoco.environments import LocoEnv
from loco_mujoco.utils import check_validity_task_mode_dataset


class MyoSkeleton(LocoEnv):
    """
    Description
    ------------

    MuJoCo simulation of the MyoSkeleton with one position controller per joint. This is a biomechanical
    humanoid model with many degrees of freedom (151 joints) to accurately simulate human movement. The model
    extends the simple humanoid by a spine, hands with fingers, realistic knees, and feet with toes.

    .. note:: This environment is currently under active development and is subject to change!


    Tasks
    -----------------
    * **Walking**: The robot has to walk forward with a fixed speed of 1.25 m/s.
    * **Running**: Run forward with a fixed speed of 2.5 m/s.


    Dataset Types
    -----------------
    The available dataset types for this environment can be found at: :ref:`env-label`.


    Observation Space
    -----------------

    The observation space has the following properties *by default* (i.e., only obs with Disabled == False):

    | For walking and running task: :code:`(min=-inf, max=inf, dim=312, dtype=float32)`

    Some observations are **disabled by default**, but can be turned on. The detailed observation space is:

    ===== ================================================ =========== ============ ======== === ========================
    Index Description                                      Min         Max          Disabled Dim Units
    ===== ================================================ =========== ============ ======== === ========================
    0     Position of Joint pelvis_ty                      -100.0      100.0        False    1   Position [m]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    1     Position of Joint pelvis_tilt                    -1.5708     1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    2     Position of Joint pelvis_list                    -1.5708     1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    3     Position of Joint pelvis_rotation                -1.5708     1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    4     Position of Joint L5_S1_Flex_Ext                 -1.22173    0.453786     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    5     Position of Joint L5_S1_Lat_Bending              -0.436332   0.436332     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    6     Position of Joint L5_S1_axial_rotation           -0.97913    0.97913      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    7     Position of Joint L4_L5_Flex_Ext                 -0.1955     0.06353      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    8     Position of Joint L4_L5_Lat_Bending              -0.03927    0.03927      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    9     Position of Joint L4_L5_axial_rotation           -0.1759     0.1759       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    10    Position of Joint L3_L4_Flex_Ext                 -0.1955     0.06353      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    11    Position of Joint L3_L4_Lat_Bending              -0.03927    0.03927      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    12    Position of Joint L3_L4_axial_rotation           -0.1759     0.1759       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    13    Position of Joint L2_L3_Flex_Ext                 -0.1955     0.06353      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    14    Position of Joint L2_L3_Lat_Bending              -0.03927    0.03927      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    15    Position of Joint L2_L3_axial_rotation           -0.1759     0.1759       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    16    Position of Joint L1_L2_Flex_Ext                 -0.1955     0.06353      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    17    Position of Joint L1_L2_Lat_Bending              -0.03927    0.03927      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    18    Position of Joint L1_L2_axial_rotation           -0.1759     0.1759       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    19    Position of Joint L1_T12_Flex_Ext                -0.1955     0.06353      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    20    Position of Joint L1_T12_Lat_Bending             -0.03927    0.03927      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    21    Position of Joint L1_T12_axial_rotation          -0.1759     0.1759       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    22    Position of Joint c7_c6_FE                       -0.05702    0.08294      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    23    Position of Joint c7_c6_LB                       -0.04492    0.04492      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    24    Position of Joint c7_c6_AR                       -0.02969    0.02969      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    25    Position of Joint c6_c5_FE                       -0.124      0.186        False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    26    Position of Joint c6_c5_LB                       -0.05236    0.05236      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    27    Position of Joint c6_c5_AR                       -0.00995    0.00995      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    28    Position of Joint c5_c4_FE                       -0.124      0.186        False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    29    Position of Joint c5_c4_LB                       -0.05236    0.05236      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    30    Position of Joint c5_c4_AR                       -0.00854    0.00854      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    31    Position of Joint c4_c3_FE                       -0.124      0.186        False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    32    Position of Joint c4_c3_LB                       -0.05236    0.05236      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    33    Position of Joint c4_c3_AR                       -0.00854    0.00854      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    34    Position of Joint c3_c2_FE                       -0.124      0.186        False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    35    Position of Joint c3_c2_LB                       -0.05236    0.05236      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    36    Position of Joint c3_c2_AR                       -0.00854    0.00854      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    37    Position of Joint c2_c1_FE                       -0.124      0.186        False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    38    Position of Joint c2_c1_LB                       -0.05236    0.05236      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    39    Position of Joint c2_c1_AR                       -0.23771    0.23771      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    40    Position of Joint c1_skull_FE                    -0.124      0.186        False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    41    Position of Joint c1_skull_LB                    -0.05236    0.05236      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    42    Position of Joint c1_skull_AR                    -0.5896     0.5896       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    43    Position of Joint skull_FE                       -0.1552     0.2328       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    44    Position of Joint skull_LB                       -0.05236    0.05236      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    45    Position of Joint skull_AR                       -0.0736     0.0736       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    46    Position of Joint sternoclavicular_r2_r          -0.269      0.0          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    47    Position of Joint sternoclavicular_r3_r          0.0         0.322013     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    48    Position of Joint unrotscap_r3_r                 -0.116      0.0          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    49    Position of Joint unrotscap_r2_r                 0.0         0.21         False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    50    Position of Joint acromioclavicular_r2_r         -0.01       0.0          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    51    Position of Joint acromioclavicular_r3_r         0.0         0.34         False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    52    Position of Joint acromioclavicular_r1_r         0.0         0.06         False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    53    Position of Joint unrothum_r1_r                  -0.466003   0.0          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    54    Position of Joint unrothum_r3_r                  -1.03673    0.0          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    55    Position of Joint unrothum_r2_r                  0.0         0.128282     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    56    Position of Joint elv_angle_r                    -1.65806    2.26893      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    57    Position of Joint shoulder_elv_r                 0.0         3.14159      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    58    Position of Joint shoulder1_r2_r                 -2.242      1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    59    Position of Joint shoulder_rot_r                 -1.571      2.094        False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    60    Position of Joint elbow_flex_r                   0.0         2.618        False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    61    Position of Joint pro_sup                        -1.5708     1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    62    Position of Joint deviation                      -0.174533   0.436332     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    63    Position of Joint flexion_r                      -0.785398   0.785398     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    64    Position of Joint cmc_flexion_r                  -0.5        0.78         False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    65    Position of Joint cmc_abduction_r                -0.78       0.7          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    66    Position of Joint mp_flexion_r                   -0.785398   0.698132     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    67    Position of Joint ip_flexion_r                   -1.309      0.436332     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    68    Position of Joint mcp2_flexion_r                 0.0         1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    69    Position of Joint mcp2_abduction_r               -0.261799   0.261799     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    70    Position of Joint pm2_flexion_r                  0.0         1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    71    Position of Joint md2_flexion_r                  0.0         1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    72    Position of Joint mcp3_flexion_r                 0.0         1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    73    Position of Joint mcp3_abduction_r               -0.261799   0.261799     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    74    Position of Joint pm3_flexion_r                  0.0         1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    75    Position of Joint md3_flexion_r                  0.0         1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    76    Position of Joint mcp4_flexion_r                 0.0         1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    77    Position of Joint mcp4_abduction_r               -0.261799   0.261799     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    78    Position of Joint pm4_flexion_r                  0.0         1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    79    Position of Joint md4_flexion_r                  0.0         1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    80    Position of Joint mcp5_flexion_r                 0.0         1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    81    Position of Joint mcp5_abduction_r               -0.261799   0.261799     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    82    Position of Joint pm5_flexion_r                  0.0         1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    83    Position of Joint md5_flexion_r                  0.0         1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    84    Position of Joint sternoclavicular_r2_l          -0.269      0.0          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    85    Position of Joint sternoclavicular_r3_l          0.0         0.322013     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    86    Position of Joint unrotscap_r3_l                 -0.116      0.0          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    87    Position of Joint unrotscap_r2_l                 0.0         0.21         False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    88    Position of Joint acromioclavicular_r2_l         -0.01       0.0          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    89    Position of Joint acromioclavicular_r3_l         0.0         0.34         False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    90    Position of Joint acromioclavicular_r1_l         0.0         0.06         False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    91    Position of Joint unrothum_r1_l                  -0.466003   0.0          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    92    Position of Joint unrothum_r3_l                  -1.03673    0.0          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    93    Position of Joint unrothum_r2_l                  0.0         0.128282     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    94    Position of Joint elv_angle_l                    -1.65806    2.26893      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    95    Position of Joint shoulder_elv_l                 0.0         3.14159      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    96    Position of Joint shoulder1_r2_l                 -2.242      1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    97    Position of Joint shoulder_rot_l                 -1.5708     2.0944       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    98    Position of Joint elbow_flex_l                   0.0         2.26893      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    99    Position of Joint pro_sup_l                      -1.5708     1.7708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    100   Position of Joint deviation_l                    -0.785398   0.785398     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    101   Position of Joint flexion_l                      -0.785398   0.785398     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    102   Position of Joint cmc_flexion_l                  -0.5        0.78         False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    103   Position of Joint cmc_abduction_l                -0.78       0.7          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    104   Position of Joint mp_flexion_l                   -0.780796   0.0          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    105   Position of Joint ip_flexion_l                   -1.22173    0.0          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    106   Position of Joint mcp2_flexion_l                 -0.349066   1.74533      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    107   Position of Joint mcp2_abduction_l               -0.790659   0.349066     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    108   Position of Joint pm2_flexion_l                  0.0         2.0944       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    109   Position of Joint md2_flexion_l                  -0.349066   1.74533      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    110   Position of Joint mcp3_flexion_l                 -0.872665   1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    111   Position of Joint mcp3_abduction_l               -0.749066   0.349066     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    112   Position of Joint pm3_flexion_l                  0.0         2.0944       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    113   Position of Joint md3_flexion_l                  -0.349066   1.74533      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    114   Position of Joint mcp4_flexion_l                 -0.872665   1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    115   Position of Joint mcp4_abduction_l               -0.749066   0.349066     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    116   Position of Joint pm4_flexion_l                  0.0         2.0944       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    117   Position of Joint md4_flexion_l                  -0.349066   1.74533      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    118   Position of Joint mcp5_flexion_l                 -0.872665   1.5708       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    119   Position of Joint mcp5_abduction_l               -0.4        0.349066     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    120   Position of Joint pm5_flexion_l                  0.0         2.0944       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    121   Position of Joint md5_flexion_l                  -0.349066   1.74533      False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    122   Position of Joint Abs_t1                         -inf        inf          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    123   Position of Joint Abs_t2                         -inf        inf          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    124   Position of Joint Abs_r3                         -inf        inf          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    125   Position of Joint Abs_r2                         -inf        inf          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    126   Position of Joint Abs_r1                         -inf        inf          False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    127   Position of Joint hip_flexion_r                  -0.523599   2.0944       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    128   Position of Joint hip_adduction_r                -0.872665   0.523599     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    129   Position of Joint hip_rotation_r                 -0.698132   0.698132     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    130   Position of Joint knee_angle_r_translation2      7.69254e-11 0.006792     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    131   Position of Joint knee_angle_r_translation1      9.53733e-08 0.00159883   False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    132   Position of Joint knee_angle_r                   0.0         2.0944       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    133   Position of Joint knee_angle_r_rotation2         -0.00167821 0.0335354    False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    134   Position of Joint knee_angle_r_rotation3         1.08939e-08 0.262788     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    135   Position of Joint ankle_angle_r                  -0.698132   0.523599     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    136   Position of Joint subtalar_angle_r               -0.349066   0.349066     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    137   Position of Joint mtp_angle_r                    -0.523599   0.523599     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    138   Position of Joint knee_angle_r_beta_translation2 -0.0408267  -0.0108281   False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    139   Position of Joint knee_angle_r_beta_translation1 -0.0227731  0.0524192    False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    140   Position of Joint knee_angle_r_beta_rotation1    -1.79241    0.010506     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    141   Position of Joint hip_flexion_l                  -0.523599   2.0944       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    142   Position of Joint hip_adduction_l                -0.872665   0.523599     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    143   Position of Joint hip_rotation_l                 -0.698132   0.698132     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    144   Position of Joint knee_angle_l_translation2      -0.006792   -7.69254e-11 False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    145   Position of Joint knee_angle_l_translation1      9.53733e-08 0.00159883   False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    146   Position of Joint knee_angle_l                   0.0         2.0944       False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    147   Position of Joint knee_angle_l_rotation2         -0.00167821 0.0335354    False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    148   Position of Joint knee_angle_l_rotation3         -0.262788   -1.08939e-08 False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    149   Position of Joint ankle_angle_l                  -0.698132   0.523599     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    150   Position of Joint subtalar_angle_l               -0.349066   0.349066     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    151   Position of Joint mtp_angle_l                    -0.523599   0.523599     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    152   Position of Joint knee_angle_l_beta_translation2 -0.0408267  -0.0108281   False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    153   Position of Joint knee_angle_l_beta_translation1 -0.0227731  0.0524192    False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    154   Position of Joint knee_angle_l_beta_rotation1    -1.79241    0.010506     False    1   Angle [rad]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    155   Velocity of Joint pelvis_tx                      -inf        inf          False    1   Velocity [m/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    156   Velocity of Joint pelvis_tz                      -inf        inf          False    1   Velocity [m/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    157   Velocity of Joint pelvis_ty                      -inf        inf          False    1   Velocity [m/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    158   Velocity of Joint pelvis_tilt                    -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    159   Velocity of Joint pelvis_list                    -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    160   Velocity of Joint pelvis_rotation                -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    161   Velocity of Joint L5_S1_Flex_Ext                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    162   Velocity of Joint L5_S1_Lat_Bending              -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    163   Velocity of Joint L5_S1_axial_rotation           -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    164   Velocity of Joint L4_L5_Flex_Ext                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    165   Velocity of Joint L4_L5_Lat_Bending              -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    166   Velocity of Joint L4_L5_axial_rotation           -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    167   Velocity of Joint L3_L4_Flex_Ext                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    168   Velocity of Joint L3_L4_Lat_Bending              -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    169   Velocity of Joint L3_L4_axial_rotation           -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    170   Velocity of Joint L2_L3_Flex_Ext                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    171   Velocity of Joint L2_L3_Lat_Bending              -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    172   Velocity of Joint L2_L3_axial_rotation           -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    173   Velocity of Joint L1_L2_Flex_Ext                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    174   Velocity of Joint L1_L2_Lat_Bending              -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    175   Velocity of Joint L1_L2_axial_rotation           -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    176   Velocity of Joint L1_T12_Flex_Ext                -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    177   Velocity of Joint L1_T12_Lat_Bending             -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    178   Velocity of Joint L1_T12_axial_rotation          -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    179   Velocity of Joint c7_c6_FE                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    180   Velocity of Joint c7_c6_LB                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    181   Velocity of Joint c7_c6_AR                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    182   Velocity of Joint c6_c5_FE                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    183   Velocity of Joint c6_c5_LB                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    184   Velocity of Joint c6_c5_AR                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    185   Velocity of Joint c5_c4_FE                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    186   Velocity of Joint c5_c4_LB                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    187   Velocity of Joint c5_c4_AR                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    188   Velocity of Joint c4_c3_FE                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    189   Velocity of Joint c4_c3_LB                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    190   Velocity of Joint c4_c3_AR                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    191   Velocity of Joint c3_c2_FE                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    192   Velocity of Joint c3_c2_LB                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    193   Velocity of Joint c3_c2_AR                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    194   Velocity of Joint c2_c1_FE                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    195   Velocity of Joint c2_c1_LB                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    196   Velocity of Joint c2_c1_AR                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    197   Velocity of Joint c1_skull_FE                    -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    198   Velocity of Joint c1_skull_LB                    -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    199   Velocity of Joint c1_skull_AR                    -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    200   Velocity of Joint skull_FE                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    201   Velocity of Joint skull_LB                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    202   Velocity of Joint skull_AR                       -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    203   Velocity of Joint sternoclavicular_r2_r          -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    204   Velocity of Joint sternoclavicular_r3_r          -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    205   Velocity of Joint unrotscap_r3_r                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    206   Velocity of Joint unrotscap_r2_r                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    207   Velocity of Joint acromioclavicular_r2_r         -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    208   Velocity of Joint acromioclavicular_r3_r         -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    209   Velocity of Joint acromioclavicular_r1_r         -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    210   Velocity of Joint unrothum_r1_r                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    211   Velocity of Joint unrothum_r3_r                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    212   Velocity of Joint unrothum_r2_r                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    213   Velocity of Joint elv_angle_r                    -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    214   Velocity of Joint shoulder_elv_r                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    215   Velocity of Joint shoulder1_r2_r                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    216   Velocity of Joint shoulder_rot_r                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    217   Velocity of Joint elbow_flex_r                   -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    218   Velocity of Joint pro_sup                        -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    219   Velocity of Joint deviation                      -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    220   Velocity of Joint flexion_r                      -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    221   Velocity of Joint cmc_flexion_r                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    222   Velocity of Joint cmc_abduction_r                -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    223   Velocity of Joint mp_flexion_r                   -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    224   Velocity of Joint ip_flexion_r                   -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    225   Velocity of Joint mcp2_flexion_r                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    226   Velocity of Joint mcp2_abduction_r               -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    227   Velocity of Joint pm2_flexion_r                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    228   Velocity of Joint md2_flexion_r                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    229   Velocity of Joint mcp3_flexion_r                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    230   Velocity of Joint mcp3_abduction_r               -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    231   Velocity of Joint pm3_flexion_r                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    232   Velocity of Joint md3_flexion_r                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    233   Velocity of Joint mcp4_flexion_r                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    234   Velocity of Joint mcp4_abduction_r               -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    235   Velocity of Joint pm4_flexion_r                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    236   Velocity of Joint md4_flexion_r                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    237   Velocity of Joint mcp5_flexion_r                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    238   Velocity of Joint mcp5_abduction_r               -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    239   Velocity of Joint pm5_flexion_r                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    240   Velocity of Joint md5_flexion_r                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    241   Velocity of Joint sternoclavicular_r2_l          -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    242   Velocity of Joint sternoclavicular_r3_l          -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    243   Velocity of Joint unrotscap_r3_l                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    244   Velocity of Joint unrotscap_r2_l                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    245   Velocity of Joint acromioclavicular_r2_l         -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    246   Velocity of Joint acromioclavicular_r3_l         -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    247   Velocity of Joint acromioclavicular_r1_l         -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    248   Velocity of Joint unrothum_r1_l                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    249   Velocity of Joint unrothum_r3_l                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    250   Velocity of Joint unrothum_r2_l                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    251   Velocity of Joint elv_angle_l                    -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    252   Velocity of Joint shoulder_elv_l                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    253   Velocity of Joint shoulder1_r2_l                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    254   Velocity of Joint shoulder_rot_l                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    255   Velocity of Joint elbow_flex_l                   -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    256   Velocity of Joint pro_sup_l                      -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    257   Velocity of Joint deviation_l                    -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    258   Velocity of Joint flexion_l                      -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    259   Velocity of Joint cmc_flexion_l                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    260   Velocity of Joint cmc_abduction_l                -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    261   Velocity of Joint mp_flexion_l                   -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    262   Velocity of Joint ip_flexion_l                   -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    263   Velocity of Joint mcp2_flexion_l                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    264   Velocity of Joint mcp2_abduction_l               -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    265   Velocity of Joint pm2_flexion_l                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    266   Velocity of Joint md2_flexion_l                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    267   Velocity of Joint mcp3_flexion_l                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    268   Velocity of Joint mcp3_abduction_l               -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    269   Velocity of Joint pm3_flexion_l                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    270   Velocity of Joint md3_flexion_l                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    271   Velocity of Joint mcp4_flexion_l                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    272   Velocity of Joint mcp4_abduction_l               -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    273   Velocity of Joint pm4_flexion_l                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    274   Velocity of Joint md4_flexion_l                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    275   Velocity of Joint mcp5_flexion_l                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    276   Velocity of Joint mcp5_abduction_l               -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    277   Velocity of Joint pm5_flexion_l                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    278   Velocity of Joint md5_flexion_l                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    279   Velocity of Joint Abs_t1                         -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    280   Velocity of Joint Abs_t2                         -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    281   Velocity of Joint Abs_r3                         -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    282   Velocity of Joint Abs_r2                         -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    283   Velocity of Joint Abs_r1                         -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    284   Velocity of Joint hip_flexion_r                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    285   Velocity of Joint hip_adduction_r                -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    286   Velocity of Joint hip_rotation_r                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    287   Velocity of Joint knee_angle_r_translation2      -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    288   Velocity of Joint knee_angle_r_translation1      -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    289   Velocity of Joint knee_angle_r                   -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    290   Velocity of Joint knee_angle_r_rotation2         -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    291   Velocity of Joint knee_angle_r_rotation3         -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    292   Velocity of Joint ankle_angle_r                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    293   Velocity of Joint subtalar_angle_r               -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    294   Velocity of Joint mtp_angle_r                    -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    295   Velocity of Joint knee_angle_r_beta_translation2 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    296   Velocity of Joint knee_angle_r_beta_translation1 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    297   Velocity of Joint knee_angle_r_beta_rotation1    -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    298   Velocity of Joint hip_flexion_l                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    299   Velocity of Joint hip_adduction_l                -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    300   Velocity of Joint hip_rotation_l                 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    301   Velocity of Joint knee_angle_l_translation2      -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    302   Velocity of Joint knee_angle_l_translation1      -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    303   Velocity of Joint knee_angle_l                   -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    304   Velocity of Joint knee_angle_l_rotation2         -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    305   Velocity of Joint knee_angle_l_rotation3         -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    306   Velocity of Joint ankle_angle_l                  -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    307   Velocity of Joint subtalar_angle_l               -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    308   Velocity of Joint mtp_angle_l                    -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    309   Velocity of Joint knee_angle_l_beta_translation2 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    310   Velocity of Joint knee_angle_l_beta_translation1 -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    311   Velocity of Joint knee_angle_l_beta_rotation1    -inf        inf          False    1   Angular Velocity [rad/s]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    312   5x3D linear Forces between Right Foot and Floor  0.0         inf          True     15  Force [N]
    ----- ------------------------------------------------ ----------- ------------ -------- --- ------------------------
    327   5X3D linear Forces between Left Foot and Floor   0.0         inf          True     15  Force [N]
    ===== ================================================ =========== ============ ======== === ========================

    Action Space
    ------------

    | The action space has the following properties *by default* (i.e., only actions with Disabled == False):
    | :code:`(min=-1, max=1, dim=151, dtype=float32)`

    The action range in LocoMuJoCo is always standardized, i.e. in [-1.0, 1.0].
    This environment is using a normalized position controller.

    Some actions can be **disabled by default**, but can be also turned on. The detailed action space is:

    ===== ================================== =========== =========== ========
    Index Name in XML                        Control Min Control Max Disabled
    ===== ================================== =========== =========== ========
    0     act_L5_S1_Flex_Ext                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    1     act_L5_S1_Lat_Bending              -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    2     act_L5_S1_axial_rotation           -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    3     act_L4_L5_Flex_Ext                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    4     act_L4_L5_Lat_Bending              -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    5     act_L4_L5_axial_rotation           -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    6     act_L3_L4_Flex_Ext                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    7     act_L3_L4_Lat_Bending              -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    8     act_L3_L4_axial_rotation           -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    9     act_L2_L3_Flex_Ext                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    10    act_L2_L3_Lat_Bending              -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    11    act_L2_L3_axial_rotation           -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    12    act_L1_L2_Flex_Ext                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    13    act_L1_L2_Lat_Bending              -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    14    act_L1_L2_axial_rotation           -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    15    act_L1_T12_Flex_Ext                -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    16    act_L1_T12_Lat_Bending             -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    17    act_L1_T12_axial_rotation          -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    18    act_c7_c6_FE                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    19    act_c7_c6_LB                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    20    act_c7_c6_AR                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    21    act_c6_c5_FE                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    22    act_c6_c5_LB                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    23    act_c6_c5_AR                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    24    act_c5_c4_FE                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    25    act_c5_c4_LB                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    26    act_c5_c4_AR                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    27    act_c4_c3_FE                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    28    act_c4_c3_LB                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    29    act_c4_c3_AR                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    30    act_c3_c2_FE                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    31    act_c3_c2_LB                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    32    act_c3_c2_AR                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    33    act_c2_c1_FE                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    34    act_c2_c1_LB                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    35    act_c2_c1_AR                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    36    act_c1_skull_FE                    -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    37    act_c1_skull_LB                    -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    38    act_c1_skull_AR                    -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    39    act_skull_FE                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    40    act_skull_LB                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    41    act_skull_AR                       -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    42    act_sternoclavicular_r2_r          -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    43    act_sternoclavicular_r3_r          -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    44    act_unrotscap_r3_r                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    45    act_unrotscap_r2_r                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    46    act_acromioclavicular_r2_r         -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    47    act_acromioclavicular_r3_r         -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    48    act_acromioclavicular_r1_r         -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    49    act_unrothum_r1_r                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    50    act_unrothum_r3_r                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    51    act_unrothum_r2_r                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    52    act_elv_angle_r                    -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    53    act_shoulder_elv_r                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    54    act_shoulder1_r2_r                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    55    act_shoulder_rot_r                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    56    act_elbow_flex_r                   -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    57    act_pro_sup                        -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    58    act_deviation                      -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    59    act_flexion_r                      -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    60    act_cmc_flexion_r                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    61    act_cmc_abduction_r                -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    62    act_mp_flexion_r                   -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    63    act_ip_flexion_r                   -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    64    act_mcp2_flexion_r                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    65    act_mcp2_abduction_r               -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    66    act_pm2_flexion_r                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    67    act_md2_flexion_r                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    68    act_mcp3_flexion_r                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    69    act_mcp3_abduction_r               -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    70    act_pm3_flexion_r                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    71    act_md3_flexion_r                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    72    act_mcp4_flexion_r                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    73    act_mcp4_abduction_r               -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    74    act_pm4_flexion_r                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    75    act_md4_flexion_r                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    76    act_mcp5_flexion_r                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    77    act_mcp5_abduction_r               -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    78    act_pm5_flexion_r                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    79    act_md5_flexion_r                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    80    act_sternoclavicular_r2_l          -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    81    act_sternoclavicular_r3_l          -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    82    act_unrotscap_r3_l                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    83    act_unrotscap_r2_l                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    84    act_acromioclavicular_r2_l         -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    85    act_acromioclavicular_r3_l         -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    86    act_acromioclavicular_r1_l         -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    87    act_unrothum_r1_l                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    88    act_unrothum_r3_l                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    89    act_unrothum_r2_l                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    90    act_elv_angle_l                    -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    91    act_shoulder_elv_l                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    92    act_shoulder1_r2_l                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    93    act_shoulder_rot_l                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    94    act_elbow_flex_l                   -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    95    act_pro_sup_l                      -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    96    act_deviation_l                    -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    97    act_flexion_l                      -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    98    act_cmc_flexion_l                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    99    act_cmc_abduction_l                -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    100   act_mp_flexion_l                   -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    101   act_ip_flexion_l                   -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    102   act_mcp2_flexion_l                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    103   act_mcp2_abduction_l               -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    104   act_pm2_flexion_l                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    105   act_md2_flexion_l                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    106   act_mcp3_flexion_l                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    107   act_mcp3_abduction_l               -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    108   act_pm3_flexion_l                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    109   act_md3_flexion_l                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    110   act_mcp4_flexion_l                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    111   act_mcp4_abduction_l               -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    112   act_pm4_flexion_l                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    113   act_md4_flexion_l                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    114   act_mcp5_flexion_l                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    115   act_mcp5_abduction_l               -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    116   act_pm5_flexion_l                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    117   act_md5_flexion_l                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    118   act_Abs_t1                         -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    119   act_Abs_t2                         -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    120   act_Abs_r3                         -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    121   act_Abs_r2                         -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    122   act_Abs_r1                         -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    123   act_hip_flexion_r                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    124   act_hip_adduction_r                -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    125   act_hip_rotation_r                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    126   act_knee_angle_r_translation2      -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    127   act_knee_angle_r_translation1      -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    128   act_knee_angle_r                   -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    129   act_knee_angle_r_rotation2         -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    130   act_knee_angle_r_rotation3         -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    131   act_ankle_angle_r                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    132   act_subtalar_angle_r               -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    133   act_mtp_angle_r                    -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    134   act_knee_angle_r_beta_translation2 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    135   act_knee_angle_r_beta_translation1 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    136   act_knee_angle_r_beta_rotation1    -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    137   act_hip_flexion_l                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    138   act_hip_adduction_l                -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    139   act_hip_rotation_l                 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    140   act_knee_angle_l_translation2      -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    141   act_knee_angle_l_translation1      -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    142   act_knee_angle_l                   -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    143   act_knee_angle_l_rotation2         -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    144   act_knee_angle_l_rotation3         -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    145   act_ankle_angle_l                  -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    146   act_subtalar_angle_l               -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    147   act_mtp_angle_l                    -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    148   act_knee_angle_l_beta_translation2 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    149   act_knee_angle_l_beta_translation1 -1.0        1.0         False
    ----- ---------------------------------- ----------- ----------- --------
    150   act_knee_angle_l_beta_rotation1    -1.0        1.0         False
    ===== ================================== =========== =========== ========

    Rewards
    --------

    The default reward function is based on the distance between the current center of mass velocity and the
    desired velocity in the x-axis. The desired velocity is given by the dataset to imitate.

    **Class**: :class:`loco_mujoco.utils.reward.TargetVelocityReward`

    Initial States
    ---------------

    The initial state is sampled by default from the dataset to imitate.

    Terminal States
    ----------------

    The terminal state is reached when the humanoid falls, or rather starts falling. The condition to check if the humanoid
    is falling is based on the orientation of the humanoid, the height of the center of mass, and the orientation of the
    back joint. More details can be found in the  :code:`_has_fallen` method of the environment.

    Methods
    ------------

    """

    valid_task_confs = ValidTaskConf(tasks=["walk", "run"],
                                     data_types=["real"])

    def __init__(self, **kwargs):
        """
        Constructor.

        """

        xml_path = (Path(__file__).resolve().parent.parent / "data" / "myo_model" /
                    "myoskeleton" / "myoskeleton.xml").as_posix()

        # check if file exists, if not exit
        if not os.path.exists(xml_path):
            print("MyoSkeleton model not initialized. Please run \"loco-mujoco-myomodel-init\" to accept the license "
                  "and download the model. Exiting...")
            exit()

        # create temporary xml file path
        xml_path_tmp = (Path(__file__).resolve().parent.parent / "data" / "myo_model_tmp" /
                        "myoskeleton" / "myoskeleton.xml").as_posix()

        # make changes to the xml to make it mjcf compatible
        xml_handle = self._apply_xml_changes(xml_path, xml_path_tmp)

        # replace the free joint with the 6 distinct joints
        xml_handle = self._add_root_joints(xml_handle)

        # add actuators
        xml_handle = self._add_actuators(xml_handle)

        # save xml_handle
        self._xml_handles = [xml_handle]

        action_spec = self._get_action_specification()

        observation_spec = self._get_observation_specification()

        self._hidable_obs = ("positions", "velocities", "foot_forces", "weight")

        _, _, _, collision_groups = self._get_xml_modifications()

        super().__init__(xml_handle, action_spec, observation_spec, collision_groups, **kwargs)

    def _get_ground_forces(self):
        """
        Returns the ground forces (np.array).

        """

        grf = np.concatenate([self._get_collision_force("floor", "right_foot1")[:3],
                              self._get_collision_force("floor", "right_foot2")[:3],
                              self._get_collision_force("floor", "right_foot3")[:3],
                              self._get_collision_force("floor", "right_foot4")[:3],
                              self._get_collision_force("floor", "right_foot5")[:3],
                              self._get_collision_force("floor", "left_foot1")[:3],
                              self._get_collision_force("floor", "left_foot2")[:3],
                              self._get_collision_force("floor", "left_foot3")[:3],
                              self._get_collision_force("floor", "left_foot4")[:3],
                              self._get_collision_force("floor", "left_foot5")[:3]]
                             )

        return grf

    def _get_xml_modifications(self):
        """
        Function that specifies which joints, motors and equality constraints
        should be removed from the Mujoco xml. Also the required collision
        groups will be returned.

        Returns:
            A tuple of lists consisting of names of joints to remove, names of motors to remove,
             names of equality constraints to remove, and names of collision groups to be used.

        """

        # for now, no modifications are implemented
        joints_to_remove = []
        motors_to_remove = []
        equ_constr_to_remove = []

        collision_groups = [("floor", ["floor"]),
                            ("right_foot1", ["foot1_r_coll"]),
                            ("right_foot2", ["foot2_r_coll"]),
                            ("right_foot3", ["foot3_r_coll"]),
                            ("right_foot4", ["bofoot1_r_coll"]),
                            ("right_foot5", ["bofoot2_r_coll"]),
                            ("left_foot1", ["foot1_l_coll"]),
                            ("left_foot2", ["foot2_l_coll"]),
                            ("left_foot3", ["foot3_l_coll"]),
                            ("left_foot4", ["bofoot1_l_coll"]),
                            ("left_foot5", ["bofoot2_l_coll"])]

        return joints_to_remove, motors_to_remove, equ_constr_to_remove, collision_groups

    def _has_fallen(self, obs, return_err_msg=False):
        """
        Checks if a model has fallen.

        Args:
            obs (np.array): Current observation.
            return_err_msg (bool): If True, an error message with violations is returned.

        Returns:
            True, if the model has fallen for the current observation, False otherwise.
            Optionally an error message is returned.

        """

        pelvis_euler = self._get_from_obs(obs, ["q_pelvis_tilt", "q_pelvis_list", "q_pelvis_rotation"])

        pelvis_height_condition = (obs[0] < -0.46) or (obs[0] > 0.1)
        pelvis_tilt_condition = (pelvis_euler[0] < (-np.pi / 4.5)) or (pelvis_euler[0] > (np.pi / 12))
        pelvis_list_condition = (pelvis_euler[1] < -np.pi / 12) or (pelvis_euler[1] > np.pi / 8)
        pelvis_rotation_condition = (pelvis_euler[2] < (-np.pi / 9)) or (pelvis_euler[2] > (np.pi / 9))

        pelvis_condition = (pelvis_height_condition or pelvis_tilt_condition
                            or pelvis_list_condition or pelvis_rotation_condition)

        lumbar_euler = self._get_from_obs(obs, ["q_L5_S1_Flex_Ext", "q_L5_S1_Lat_Bending", "q_L5_S1_axial_rotation"])

        lumbar_extension_condition = (lumbar_euler[0] < (-np.pi / 4)) or (lumbar_euler[0] > (np.pi / 10))
        lumbar_bending_condition = (lumbar_euler[1] < -np.pi / 10) or (lumbar_euler[1] > np.pi / 10)
        lumbar_rotation_condition = (lumbar_euler[2] < (-np.pi / 4.5)) or (lumbar_euler[2] > (np.pi / 4.5))

        lumbar_condition = (lumbar_extension_condition or lumbar_bending_condition or lumbar_rotation_condition)

        if return_err_msg:
            error_msg = ""
            if pelvis_height_condition:
                error_msg += "pelvis_height_condition violated.\n"
            if pelvis_tilt_condition:
                error_msg += "pelvis_tilt_condition violated.\n"
            if pelvis_list_condition:
                error_msg += "pelvis_list_condition violated.\n"
            if pelvis_rotation_condition:
                error_msg += "pelvis_rotation_condition violated.\n"
            if lumbar_extension_condition:
                error_msg += "lumbar_extension_condition violated.\n"
            if lumbar_bending_condition:
                error_msg += "lumbar_bending_condition violated.\n"
            if lumbar_rotation_condition:
                error_msg += "lumbar_rotation_condition violated.\n"

            return pelvis_condition or lumbar_condition, error_msg
        else:
            return (pelvis_condition or lumbar_condition)

    def _get_observation_specification(self):
        """
        Getter for the observation space specification. This function reads all joint names from the xml and adds
        the prefix "q_" for the joint positions and "dq_" for the joint velocities.

        Returns:
            A list of tuples containing the specification of each observation
            space entry.

        """
        observation_spec = []
        for prefix in ["q_", "dq_"]:
            for j in self.xml_handle.find_all("joint"):
                obs_type = ObservationType.JOINT_POS if prefix == "q_" else ObservationType.JOINT_VEL
                observation_spec.append((prefix + j.name, j.name, obs_type))
        return observation_spec

    def _get_action_specification(self):
        """
        Getter for the action space specification. This function add alls actuator names found in the xml, which
        are the ones added in the _add_actuators method.

        Returns:
            A list of tuples containing the specification of each action
            space entry.

        """
        action_spec = []
        actuators = self.xml_handle.find_all("actuator")
        for actuator in actuators:
            action_spec.append(actuator.name)
        return action_spec

    @staticmethod
    def generate(task="walk", dataset_type="real", debug=False, **kwargs):
        """
        Returns a Full-body MyoSkeleton environment and a dataset corresponding to the specified task.

        Args:
            task (str): Main task to solve. Either "walk" or "run".
            dataset_type (str): "real" or "perfect". "real" uses real motion capture data as the
                reference trajectory. This data does not perfectly match the kinematics
                and dynamics of this environment, hence it is more challenging. "perfect" uses
                a perfect dataset.
            debug (bool): If True, the smaller test datasets are used for debugging purposes.


        Returns:
            An MDP of a Full-body MyoSkeleton Humanoid.

        """

        check_validity_task_mode_dataset(MyoSkeleton.__name__, task, None, dataset_type,
                                         *MyoSkeleton.valid_task_confs.get_all())

        if "reward_type" in kwargs.keys():
            reward_type = kwargs["reward_type"]
            del kwargs["reward_type"]
        else:
            reward_type = "target_velocity"

        if task == "walk":
            path = "datasets/humanoids/real/myosuite_humanoid_walking.npz"
            use_mini_dataset = not os.path.exists(Path(loco_mujoco.__file__).resolve().parent / path)
            if debug or use_mini_dataset:
                if use_mini_dataset:
                    raise NotImplementedError("Mini datasets are currently not available for the MyoSkeleton.")
                path = path.split("/")
                path.insert(3, "mini_datasets")
                path = "/".join(path)
            traj_path = Path(loco_mujoco.__file__).resolve().parent / path
            if "reward_params" in kwargs.keys():
                reward_params = kwargs["reward_params"]
                del kwargs["reward_params"]
            else:
                reward_params = dict(target_velocity=1.25)
        elif task == "run":
            path = "datasets/humanoids/real/myosuite_humanoid_running.npz"
            use_mini_dataset = not os.path.exists(Path(loco_mujoco.__file__).resolve().parent / path)
            if debug or use_mini_dataset:
                if use_mini_dataset:
                    raise NotImplementedError("Mini datasets are currently not available for the MyoSkeleton.")
                path = path.split("/")
                path.insert(3, "mini_datasets")
                path = "/".join(path)
            traj_path = Path(loco_mujoco.__file__).resolve().parent / path
            if "reward_params" in kwargs.keys():
                reward_params = kwargs["reward_params"]
                del kwargs["reward_params"]
            else:
                reward_params = dict(target_velocity=2.5)

        # Generate the MDP
        mdp = MyoSkeleton(reward_type=reward_type, reward_params=reward_params, **kwargs)

        # Load the trajectory
        env_freq = 1 / mdp._timestep  # hz
        desired_contr_freq = 1 / mdp.dt  # hz
        n_substeps = env_freq // desired_contr_freq

        if dataset_type == "real":
            traj_data_freq = 500  # hz
            traj_params = dict(traj_path=traj_path,
                               traj_dt=(1 / traj_data_freq),
                               control_dt=(1 / desired_contr_freq))
        else:
            raise ValueError(f"currently not implemented.")

        mdp.load_trajectory(traj_params, warn=False)

        return mdp

    @staticmethod
    def _apply_xml_changes(xml_path, xml_path_tmp):
        """
        This function reads the original myo_model xml and applies some changes to make it usable in LocoMuJoCo.
        It creates a temporary xml file with the changes applied.

        Args:
            xml_path (str): Path to the original xml.
            xml_path_tmp (str): Path to the tmp xml to be created.

        Returns:
            Modified Mujoco XML Handle.

        """

        # Create tmp dir if it doesn't exist yet
        os.makedirs(os.path.dirname(xml_path_tmp), exist_ok=True)

        # we load and save the model to have a single file xml
        model = mujoco.MjModel.from_xml_path(xml_path)
        mujoco.mj_saveLastXML(xml_path_tmp, model)

        # load original xml
        tree = ET.parse(xml_path_tmp)
        root = tree.getroot()

        def _get_parent(elem):
            """
            Get the parent of an XML element.
            """
            return next((p for p in root.iter() if elem in list(p)), None)

        # rename all file paths of textures and meshes
        assets = root.find('asset')
        for e in assets:
            if e.tag == "mesh" or e.tag == "texture":
                if "file" in e.attrib.keys():
                    test = e.attrib["file"]
                    file_name = test.split("/")
                    e.attrib["file"] = "/".join(file_name)

        # delete original cameras, floor and lighting added to the world body.
        wb = root.find('worldbody')
        elems_to_delete = []
        for e in wb:
            if e.tag != "body":
                elems_to_delete.append(e)
        for e in elems_to_delete:
            wb.remove(e)

        # Find and delete all frame elements (not mjcf parseable)
        frames_to_remove = root.findall('.//frame')

        for frame in frames_to_remove:
            # Find the parent of the frame element by iterating over the tree
            parent = _get_parent(frame)
            if parent is not None:
                # Move all children of the frame to its parent
                children = list(frame)
                for child in children:
                    parent.append(child)
                # Remove the frame from its parent
                parent.remove(frame)

        # remove free joint
        for joint in root.findall('.//joint'):
            if "name" in joint.attrib.keys():
                if joint.attrib["name"] == "myoskeleton_root":
                    parent = _get_parent(joint)
                    parent.remove(joint)
                    break

        # Add LocoMuJoCo default elements
        # add floor
        ET.SubElement(wb, "geom", {"name": "floor", "pos": "0 0 0", "size": "200 200 0.125", "type": "plane",
                                   "material": "MatPlane", "condim": "3", "group": "2", "rgba": "0.8 0.9 0.8 1"})
        # add light
        ET.SubElement(wb, "light", {"cutoff": "100", "diffuse": "1 1 1", "dir": "-0 0 -1.3",
                                    "directional": "true", "exponent": "1", "pos": "0 0 1.3",
                                    "specular": ".1 .1 .1", "castshadow": "false"})

        # add materials and textures
        ET.SubElement(assets, "texture", {"name": "texplane", "type": "2d", "builtin": "checker", "rgb1": ".2 .3 .4",
                                          "rgb2": ".1 0.15 0.2", "width": "100", "height": "100"})
        ET.SubElement(assets, "material", {"name": "MatPlane", "reflectance": "0.5", "texture": "texplane",
                                           "texrepeat": "1 1", "texuniform": "true"})
        ET.SubElement(assets, "texture", {"builtin": "gradient", "height": "100", "rgb1": ".4 .5 .6", "rgb2": "0 0 0",
                                          "type": "skybox", "width": "100"})

        # write temporary xml and load xml_handle in mjcf
        tree.write(xml_path_tmp)

        # load from xml handle
        xml_handle = mjcf.from_path(xml_path_tmp)

        return xml_handle

    @staticmethod
    def _add_root_joints(xml_handle):
        """
        Adds 6 distinct degrees of freedom to the pelvis to  align with the
        LocoMuJoCo root joint convention.

        """

        # find root and add 6 joints replacing the free joint
        pelvis_body = xml_handle.find("body", "pelvis")
        pelvis_body.quat = [0.7071067811865475, 0.7071067811865475, 0.0, 0.0]
        pelvis_body.insert("joint", position=0, name="pelvis_tx", pos=[0.0, 0.0, 0.0],
                           axis=[1, 0, 0], type="slide", range=[-500, 500])
        pelvis_body.insert("joint", position=1, name="pelvis_tz", pos=[0.0, 0.0, 0.0],
                           axis=[0, 0, 1], type="slide", range=[-500, 500])
        pelvis_body.insert("joint", position=2, name="pelvis_ty", pos=[0.0, 0.0, 0.0],
                           axis=[0, 1, 0], type="slide", range=[-100, 100])
        pelvis_body.insert("joint", position=3, name="pelvis_tilt", pos=[0.0, 0.0, 0.0],
                           axis=[0, 0, 1], range=[-1.5708, 1.5708])
        pelvis_body.insert("joint", position=4, name="pelvis_list", pos=[0.0, 0.0, 0.0],
                           axis=[1, 0, 0], range=[-1.5708, 1.5708])
        pelvis_body.insert("joint", position=5, name="pelvis_rotation", pos=[0.0, 0.0, 0.0],
                           axis=[0, 1, 0], range=[-1.5708, 1.5708])

        return xml_handle

    @staticmethod
    def _add_actuators(xml_handle):
        """
        Adds a generic actuator to each joint.

        """
        max_joint_forces = dict(L5_S1_Flex_Ext=200,
                                L5_S1_Lat_Bending=200,
                                L5_S1_axial_rotation=200,
                                L4_L5_Flex_Ext=200,
                                L4_L5_Lat_Bending=200,
                                L4_L5_axial_rotation=200,
                                L3_L4_Flex_Ext=200,
                                L3_L4_Lat_Bending=200,
                                L3_L4_axial_rotation=200,
                                L2_L3_Flex_Ext=200,
                                L2_L3_Lat_Bending=200,
                                L2_L3_axial_rotation=200,
                                L1_L2_Flex_Ext=200,
                                L1_L2_Lat_Bending=200,
                                L1_L2_axial_rotation=200,
                                L1_T12_Flex_Ext=200,
                                L1_T12_Lat_Bending=200,
                                L1_T12_axial_rotation=200,
                                c7_c6_FE=50,
                                c7_c6_LB=50,
                                c7_c6_AR=50,
                                c6_c5_FE=50,
                                c6_c5_LB=50,
                                c6_c5_AR=50,
                                c5_c4_FE=50,
                                c5_c4_LB=50,
                                c5_c4_AR=50,
                                c4_c3_FE=50,
                                c4_c3_LB=50,
                                c4_c3_AR=50,
                                c3_c2_FE=50,
                                c3_c2_LB=50,
                                c3_c2_AR=50,
                                c2_c1_FE=50,
                                c2_c1_LB=50,
                                c2_c1_AR=50,
                                c1_skull_FE=50,
                                c1_skull_LB=50,
                                c1_skull_AR=50,
                                skull_FE=50,
                                skull_LB=50,
                                skull_AR=50,
                                sternoclavicular_r2_r=80,
                                sternoclavicular_r3_r=80,
                                unrotscap_r3_r=80,
                                unrotscap_r2_r=80,
                                acromioclavicular_r2_r=80,
                                acromioclavicular_r3_r=80,
                                acromioclavicular_r1_r=80,
                                unrothum_r1_r=80,
                                unrothum_r3_r=80,
                                unrothum_r2_r=80,
                                elv_angle_r=80,
                                shoulder_elv_r=80,
                                shoulder1_r2_r=80,
                                shoulder_rot_r=80,
                                elbow_flex_r=80,
                                pro_sup=80,
                                deviation=80,
                                flexion_r=80,
                                sternoclavicular_r2_l=80,
                                sternoclavicular_r3_l=80,
                                unrotscap_r3_l=80,
                                unrotscap_r2_l=80,
                                acromioclavicular_r2_l=80,
                                acromioclavicular_r3_l=80,
                                acromioclavicular_r1_l=80,
                                unrothum_r1_l=80,
                                unrothum_r3_l=80,
                                unrothum_r2_l=80,
                                elv_angle_l=80,
                                shoulder_elv_l=80,
                                shoulder1_r2_l=80,
                                shoulder_rot_l=80,
                                elbow_flex_l=80,
                                pro_sup_l=80,
                                deviation_l=80,
                                flexion_l=80,
                                hip_flexion_r=200,
                                hip_adduction_r=200,
                                hip_rotation_r=200,
                                knee_angle_r=200,
                                knee_angle_r_rotation2=20,
                                knee_angle_r_rotation3=20,
                                ankle_angle_r=200,
                                subtalar_angle_r=200,
                                mtp_angle_r=200,
                                knee_angle_r_beta_rotation1=20,
                                hip_flexion_l=200,
                                hip_adduction_l=200,
                                hip_rotation_l=200,
                                knee_angle_l=200,
                                knee_angle_l_rotation2=20,
                                knee_angle_l_rotation3=20,
                                ankle_angle_l=200,
                                subtalar_angle_l=200,
                                mtp_angle_l=200,
                                knee_angle_l_beta_rotation1=20)

        joints = xml_handle.find_all("joint")
        for joint in joints:
            # add an actuator for every joint except the pelvis
            if "pelvis" not in joint.name:
                max_force = max_joint_forces[joint.name] if joint.name in max_joint_forces.keys() else 50
                xml_handle.actuator.add("general", name="act_" + joint.name, joint=joint.name,
                                        ctrlrange=[-max_force, max_force], ctrllimited=True)
        return xml_handle

    @staticmethod
    def _get_grf_size():
        """
        Returns the size of the ground force vector.

        """

        return 30
